#!/usr/bin/perl 

use strict;
use File::Copy;
use List::Util qw(sum);

# This first part reads the mgf file and calculates the rank of each peak within each peaklist;
print "\n(1)\tReading mgf files and \n\tcalculating the rank of each peak within each peaklist.\n";

# Configuration part;
my @reporter=();
my $dev;


# Call subroutine to read configs
my (%Config, $config);
$config = &getConfig ("config.txt", \%Config);


# read all .mgf files in the current directory
my $file;
while($file=<*.mgf>) {
    die "Cannot open source file!" unless open(Fin, $file);    
    print "\t$file...";
    
    my $start="BEGIN IONS";
    my $stop="END IONS";
    my ($output, $rank, $total, @inten, @mass, @rank, %index, $index, @index);
    my $counter=1;
    my $found=0;
    
    # check if the previous output file exist then rename it to .bkp
    my $filename = "$file".'.rnk';    
    if (-e $filename) {
        print "\n\tRank file exist!! renaming to .bkp ...\n";
        move ($filename, "$filename".'.bkp');
    }
    
    while (<Fin>) {
        my $line = $_;	

        # Find block of lines to extract                                                           
        if( /$start/../$stop/ ) {
            # Start of block                                                                       
            if ( /$start/ ) {                
                open($output,'>>'.$filename) or die $!;
                chomp($line);                
                print $output $line;			# print first line BEGIN ION
            }            
            # End of block                                                                         
            elsif ( /$stop/ ) {
                
                $total = sum(@inten);		 	# calculate total intensities           
                @rank = sort{$b <=> $a } (@inten);	# do the ranking based on peak list
                
                if ($counter >1) {shift(@inten)};
		
                for (my $n=0; $n<@inten; $n++){		# calculates indices
                    my $search = $inten[$n];
                    %index;
                    @index{@rank} = (0..$#rank);
                    $index=$index{$search};
                    my $ind=$index+1;
                    my $norminten=$inten[$n]/$total;                    
		    printf $output "%f\t%e\t%d\t\n", $mass[$n], $norminten, $ind ;		#print peaklist and rank
                }
		
                print $output "END IONS\n";							# print END IONS
                $counter++;
                $found = 0;
                (@inten, @mass, @rank, %index, @index, $total, $index) = undef;
            }
            # Middle of block                                                                      
            else{										# print first line after BEGIN IONS (TITLE)                
                if($found == 0) {                           
                    print $output "\n$line";                    
                    $found=1;
                }
                else {
                    # Extract peak lists
		    if ($line !~ /[a-zA-Z]/) {        	
			(my $mass, my $inten) = split("\t");
			chomp($inten);			# get rid of \n at the end
			if ($mass) {push (@mass, $mass)};
                        if ($inten) {push (@inten, $inten)};
		    }	
                    else {
                    print $output $line;		# print rest of the header
		    }
                }
            }
        }
        # Find block of lines to extract
    }
    close ($output);
    close Fin;
}
print "\tDone.\n ";

# The second part reads the .rnk file(s) and extracts candidate precursors
print "\n(2)\tNow reading .rnk files and \n\textracting candidate precursors.\n";


my ($line, $title, $pepmass, $pepintensity, $charge, $rawscans, $rt, $scans, @neutralloss, @oxoniumloss);
my (@tol, $tolNL, $tolOL);
my ($n, @peaklist, $switch);
my (@Reporter, @reporterint, @dev, @score, $score);


while($file=<*.rnk>) {
    die "Cannot open source file!" unless open(Fin, $file);
    die "Cannot create file!" unless open(Fout,">$file".'.out');
    print "\t$file...";
    
    print Fout "\t\t\tprecursor\t\t\t\t204\t\t\t\t186\t\t\t\t168\t\t\t\t144\t\t\t\t138\t\t\t\t126",
		"\t\t\t\tneutral loss group1\t\t\t\tneutral loss group2\t\t\t\tneutral loss group3\t\t\t\tneutral loss group4",
		"\t\t\t\tGlcNAc oxonium loss group1\t\t\t\tGlcNAc oxonium loss group2\t\t\t\tGlcNAc oxonium loss group3\t\t\t\tGlcNAc oxonium loss group4\n";
    print Fout "title\tscans\trt /s\tmass\tcharge\tintensity\tOScore"."\tmass\tintensity\tdev /$config->{unit}\tscore"x14;
	print Fout "\n";

    while ($line=<Fin>){
        
        chomp ($line);              # Get rid of the trailling \n
        $line =~ s/^\s*//;          # Remove spaces at the start of the line
        $line =~ s/\s*$//;         # Remove spaces at the end of the line
        
        if ( ($line !~ /^#/) && ($line ne "")) {         # Ignore lines starting with #, blank lines 
            
            if ($line=~/^BEGIN IONS/) {# get scan header data and reset all indices and arrays
                @peaklist = (); ($switch, $n) = 0; 
                $line = <Fin>; die "no TITLE" unless ($line =~ /^TITLE=(.*)/); $title = $1; 
                $line = <Fin>; die "no PEPMASS" unless ($line =~ /^PEPMASS=(\d+\.\d+)\s+([+-]?(\d+\.\d+|\d+\.|\.\d+|\d+)([eE][+-]?\d+)?)\s+/); $pepmass = $1; $pepintensity = $2; 
		$line = <Fin>; die "no CHARGE" unless ($line =~ /^CHARGE=(\d)+/); $charge = $1; #print $charge;
                $line = <Fin>; die "no SCANS" unless ($line =~ /^SCANS=(\d+)/); $scans = $1; 
                $line = <Fin>; die "no RAWSCANS" unless ($line =~ /^RAWSCANS=(.*)/); $rawscans = $1;            
                $line = <Fin>; die "no RTINSECONDS" unless ($line =~ /^RTINSECONDS=(.*)/); $rt = $1; 
                

		# Search for reporter and losses 
		for (my $m=1; $m<5; $m++){
		    $neutralloss[$m] = ( $pepmass - $m*($config->{deltaneutralloss}/$charge) );
		    
		    #Calculate GlcNAc oxonium ion loss mass for z>1
		    if ($charge > $m) {
			$oxoniumloss[$m] = ($pepmass*$charge-$m*1.00727649-$m*$config->{deltaneutralloss})/($charge-$m);
		    } # No GlcNAc oxonium ion loss for z=1
		    elsif ($charge == $m) {
			$oxoniumloss[$m] = undef;
		    }     
		}
		
                if ($config->{unit} eq "Da") { $dev = $config->{tol} ; }
                elsif ($config->{unit} eq "ppm"){
                    for (my $i=1; $i<=14 ; $i++) {
                        if ($i<=6){
			    $tol[$i] = ( $config->{tol} * $config->{"reporter$i"} ) / ($config->{convert});    
			}			
			if ($i>=7 && $i <=10){	# $tol[7..10]=neutralloss 1..4; 
			    for (my $m=1; $m<5; $m++) {
				$tol[$i] = ( $config->{tol} * $neutralloss[$m]) / ($config->{convert});
				$i++;
			    }
			}
			if ($i>=11 && $i<=14){	# $tol[11..14]=oxoniumloss 1..4;
			    for (my $m=1; $m<5; $m++) {
				$tol[$i] = ($config->{tol} * $oxoniumloss[$m]) / ($config->{convert});
				$i++;
			    }			    
			}						
                    }
                }
                else {
                    die "Error: unknown argument";
                }
            }
            
            # extract peaklists
            if ($line=~/^(\d+\.\d+)\t(\d+\.\d+e-\d+)\t(\d+)/){
                ($peaklist[$n][1], $peaklist[$n][2], $peaklist[$n][3]) = split("\t", $line);
                $n++;
            }
            
            if ($line=~/^END IONS/){
                (@Reporter, @reporterint, @dev, @score) = undef;                
		
                for (my $n=0; $n<@peaklist; $n++){
		    for (my $i=1; $i<=14 ; $i++) {
			
			if ($i<=6) {
			    if (abs( $config->{"reporter$i"} - $peaklist[$n][1]) < $tol[$i]) {
				$switch = 1;
				$Reporter[$i] = $peaklist[$n][1];
				$reporterint[$i] = $peaklist[$n][2];
				$dev[$i] = (($peaklist[$n][1]-$config->{"reporter$i"})/$config->{"reporter$i"}*$config->{convert} );
				$score[$i] = $peaklist[$n][2]/$peaklist[$n][3];
			    }    
			}
			
			if ($i>=7 && $i<=10){
			    for (my $m=1; $m<5; $m++) {				
				if (abs($neutralloss[$m]-$peaklist[$n][1]) < $tol[$i]){ 	#$NL, $NLint, $NLdev, $NLscore
				    $switch = 1;
				    $Reporter[$i] = $peaklist[$n][1];
				    $reporterint[$i] = $peaklist[$n][2];
				    $dev[$i] = (($peaklist[$n][1]-$neutralloss[$m])/$neutralloss[$m]*$config->{convert});				    
				    $score[$i] = $peaklist[$n][2]/$peaklist[$n][3];
				}
			    }
			}
			
			if ($i>=11 && $i<=14 ){
			    for (my $m=1; $m<5; $m++) {	
				if (abs($oxoniumloss[$m]-$peaklist[$n][1]) < $tol[$i]){	#$OL, $OLint, $OLdev, $)Lscore
				    $switch = 1;
				    $Reporter[$i] = $peaklist[$n][1];
				    $reporterint[$i] = $peaklist[$n][2];
				    $dev[$i] = (($peaklist[$n][1]-$oxoniumloss[$m])/$oxoniumloss[$m]*$config->{convert});
				    $score[$i] = $peaklist[$n][2]/$peaklist[$n][3];
				}
			    }
			}
                    }  
                }
		
                if ($switch == 1){
		    my $totalscore = 0;
                    ( $score +=$_ ) for @score;		    
                    my $log10score = -log($score)/log(10);
                    print Fout "$title\t$scans\t$rt\t$pepmass\t$charge\t$pepintensity\t$log10score",
				"\t$Reporter[6]\t$reporterint[6]\t$dev[6]\t$score[6]\t$Reporter[5]\t$reporterint[5]\t$dev[5]\t$score[5]",
				"\t$Reporter[4]\t$reporterint[4]\t$dev[4]\t$score[4]\t$Reporter[3]\t$reporterint[3]\t$dev[3]\t$score[3]",
				"\t$Reporter[2]\t$reporterint[2]\t$dev[2]\t$score[2]\t$Reporter[1]\t$reporterint[1]\t$dev[1]\t$score[1]",
				"\t$Reporter[7]\t$reporterint[7]\t$dev[7]\t$score[7]\t$Reporter[8]\t$reporterint[8]\t$dev[8]\t$score[8]",
				"\t$Reporter[9]\t$reporterint[9]\t$dev[9]\t$score[9]\t$Reporter[10]\t$reporterint[10]\t$dev[10]\t$score[10]",
				"\t$Reporter[11]\t$reporterint[11]\t$dev[11]\t$score[11]\t$Reporter[12]\t$reporterint[12]\t$dev[12]\t$score[12]",
				"\t$Reporter[13]\t$reporterint[13]\t$dev[13]\t$score[13]\t$Reporter[14]\t$reporterint[14]\t$dev[14]\t$score[14]\n";                    
                }
            }
        }
    }
    print "\n\tDone.\n";
    close Fin;
    close Fout;
}

# ==================================================================================================
# Accept configuration details to define variables at the begining, and let users to modify them. 
sub getConfig {

    my ($config_line, $Name, $Value, $Config);
    
    my ($File, $Config) = @_;
    
    if (!open (CONFIGS, "$File")) {
        print "ERROR: Config file not found : $File";
        exit(0);
    }

    while (<CONFIGS>) {
        $config_line=$_;
        chop ($config_line);           # Get rid of the trailling \n
        $config_line =~ s/^\s*//;      # Remove spaces at the start of the line
        $config_line =~ s/\s*$//;      # Remove spaces at the end of the line
        if ( ($config_line !~ /^#/) && ($config_line ne "") ){    # Ignore lines starting with # and blank lines
            ($Name, $Value) = split (/=/, $config_line);          # Split each line into name value pairs
            $Config{$Name} = $Value;                              # Create a hash of the name value pairs
        }        
    }    
    return \%Config;
    close(CONFIGS);
}

